/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.filesystem;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.apache.poi.poifs.common.POIFSBigBlockSize;
import org.apache.poi.poifs.common.POIFSConstants;
import org.apache.poi.poifs.dev.POIFSViewable;
import org.apache.poi.poifs.property.DocumentProperty;
import org.apache.poi.poifs.property.Property;
import org.apache.poi.poifs.storage.BlockWritable;
import org.apache.poi.poifs.storage.DataInputBlock;
import org.apache.poi.poifs.storage.DocumentBlock;
import org.apache.poi.poifs.storage.ListManagedBlock;
import org.apache.poi.poifs.storage.RawDataBlock;
import org.apache.poi.poifs.storage.SmallDocumentBlock;
import org.apache.poi.util.HexDump;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.math.BigInteger;
import java.util.Arrays;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * This class manages a document in the POIFS filesystem.
 *
 * @author Marc Johnson (mjohnson at apache dot org)
 */
public final class POIFSDocument implements BATManaged, BlockWritable, POIFSViewable  {
	private static final DocumentBlock[] EMPTY_BIG_BLOCK_ARRAY = { };
	private static final SmallDocumentBlock[] EMPTY_SMALL_BLOCK_ARRAY = { };
	private DocumentProperty _property;
	private int _size;
	
   private final POIFSBigBlockSize _bigBigBlockSize;

	// one of these stores will be valid
	private SmallBlockStore  _small_store;
	private BigBlockStore	 _big_store;
	
		/**
	 * Constructor from large blocks
	 *
	 * @param name the name of the POIFSDocument
	 * @param blocks the big blocks making up the POIFSDocument
	 * @param length the actual length of the POIFSDocument
	 */
	public POIFSDocument(String name, RawDataBlock[] blocks, int length) throws IOException {
		_size = length;
		if(blocks.length == 0) {
		   _bigBigBlockSize = POIFSConstants.SMALLER_BIG_BLOCK_SIZE_DETAILS;
		} else {
		   _bigBigBlockSize = (blocks[0].getBigBlockSize() == POIFSConstants.SMALLER_BIG_BLOCK_SIZE ?
		         POIFSConstants.SMALLER_BIG_BLOCK_SIZE_DETAILS : 
		         POIFSConstants.LARGER_BIG_BLOCK_SIZE_DETAILS
		   );
		}
		
		_big_store = new BigBlockStore(_bigBigBlockSize, convertRawBlocksToBigBlocks(blocks));
		_property = new DocumentProperty(name, _size);
		_small_store = new SmallBlockStore(_bigBigBlockSize, EMPTY_SMALL_BLOCK_ARRAY);
		_property.setDocument(this);
	}

	// TODO - awkward typing going on here
	private static DocumentBlock[] convertRawBlocksToBigBlocks(ListManagedBlock[] blocks) throws IOException {
		DocumentBlock[] result = new DocumentBlock[blocks.length];
		for (int i = 0; i < result.length; i++) {
			result[i] = new DocumentBlock((RawDataBlock)blocks[i]);
		}
		return result;
	}
	private static SmallDocumentBlock[] convertRawBlocksToSmallBlocks(ListManagedBlock[] blocks) {
		if (blocks instanceof SmallDocumentBlock[]) {
			return (SmallDocumentBlock[]) blocks;
		}
		SmallDocumentBlock[] result = new SmallDocumentBlock[blocks.length];
		System.arraycopy(blocks, 0, result, 0, blocks.length);
		return result;
	}

	/**
	 * Constructor from small blocks
	 *
	 * @param name the name of the POIFSDocument
	 * @param blocks the small blocks making up the POIFSDocument
	 * @param length the actual length of the POIFSDocument
	 */
	public POIFSDocument(String name, SmallDocumentBlock[] blocks, int length) {
		_size = length;
		
		if(blocks.length == 0) {
		   _bigBigBlockSize = POIFSConstants.SMALLER_BIG_BLOCK_SIZE_DETAILS;
		} else {
		   _bigBigBlockSize = blocks[0].getBigBlockSize();
		}

		_big_store = new BigBlockStore(_bigBigBlockSize, EMPTY_BIG_BLOCK_ARRAY);
		_property = new DocumentProperty(name, _size);
		_small_store = new SmallBlockStore(_bigBigBlockSize, blocks);
		_property.setDocument(this);
	}

	/**
	 * Constructor from small blocks
	 *
	 * @param name the name of the POIFSDocument
	 * @param blocks the small blocks making up the POIFSDocument
	 * @param length the actual length of the POIFSDocument
	 */
	public POIFSDocument(String name, POIFSBigBlockSize bigBlockSize, ListManagedBlock[] blocks, int length) throws IOException {
		_size = length;
		_bigBigBlockSize = bigBlockSize;
		_property = new DocumentProperty(name, _size);
		_property.setDocument(this);
		if (Property.isSmall(_size)) {
			_big_store = new BigBlockStore(bigBlockSize,EMPTY_BIG_BLOCK_ARRAY);
			_small_store = new SmallBlockStore(bigBlockSize,convertRawBlocksToSmallBlocks(blocks));
		} else {
			_big_store = new BigBlockStore(bigBlockSize,convertRawBlocksToBigBlocks(blocks));
			_small_store = new SmallBlockStore(bigBlockSize,EMPTY_SMALL_BLOCK_ARRAY);
		}
	}
	public POIFSDocument(String name, ListManagedBlock[] blocks, int length) throws IOException {
	   this(name, POIFSConstants.SMALLER_BIG_BLOCK_SIZE_DETAILS, blocks, length);
	}

	/**
	 * Constructor
	 *
	 * @param name the name of the POIFSDocument
	 * @param stream the InputStream we read data from
	 */
	public POIFSDocument(String name, POIFSBigBlockSize bigBlockSize, InputStream stream) throws IOException {
		List<DocumentBlock> blocks = new ArrayList<DocumentBlock>();

		_size = 0;
		_bigBigBlockSize = bigBlockSize;
		while (true) {
			DocumentBlock block = new DocumentBlock(stream, bigBlockSize);
			int blockSize = block.size();

			if (blockSize > 0) {
				blocks.add(block);
				_size += blockSize;
			}
			if (block.partiallyRead()) {
				break;
			}
		}
		DocumentBlock[] bigBlocks = blocks.toArray(new DocumentBlock[blocks.size()]);

		_big_store = new BigBlockStore(bigBlockSize,bigBlocks);
		_property = new DocumentProperty(name, _size);
		_property.setDocument(this);
		if (_property.shouldUseSmallBlocks()) {
			_small_store = new SmallBlockStore(bigBlockSize,SmallDocumentBlock.convert(bigBlockSize,bigBlocks, _size));
			_big_store = new BigBlockStore(bigBlockSize,new DocumentBlock[0]);
		} else {
			_small_store = new SmallBlockStore(bigBlockSize,EMPTY_SMALL_BLOCK_ARRAY);
		}
	}
	public POIFSDocument(String name, InputStream stream) throws IOException {
	   this(name, POIFSConstants.SMALLER_BIG_BLOCK_SIZE_DETAILS, stream);
	}

	/**
	 * Constructor
	 *
	 * @param name the name of the POIFSDocument
	 * @param size the length of the POIFSDocument
	 * @param path the path of the POIFSDocument
	 * @param writer the writer who will eventually write the document contents
	 */
	public POIFSDocument(String name, int size, POIFSBigBlockSize bigBlockSize, POIFSDocumentPath path, POIFSWriterListener writer) {
		_size = size;
		_bigBigBlockSize = bigBlockSize;
		_property = new DocumentProperty(name, _size);
		_property.setDocument(this);
		if (_property.shouldUseSmallBlocks()) {
			_small_store = new SmallBlockStore(_bigBigBlockSize, path, name, size, writer);
			_big_store = new BigBlockStore(_bigBigBlockSize, EMPTY_BIG_BLOCK_ARRAY);
		} else {
			_small_store = new SmallBlockStore(_bigBigBlockSize, EMPTY_SMALL_BLOCK_ARRAY);
			_big_store = new BigBlockStore(_bigBigBlockSize, path, name, size, writer);
		}
	}
	public POIFSDocument(String name, int size, POIFSDocumentPath path, POIFSWriterListener writer) {
	   this(name, size, POIFSConstants.SMALLER_BIG_BLOCK_SIZE_DETAILS, path, writer);
	}

	/**
	 * @return array of SmallDocumentBlocks; may be empty, cannot be null
	 */
	public BlockWritable[] getSmallBlocks() {
		return _small_store.getBlocks();
	}

	/**
	 * @return size of the document
	 */
	public int getSize() {
		return _size;
	}

	/**
	 * read data from the internal stores
	 *
	 * @param buffer the buffer to write to
	 * @param offset the offset into our storage to read from
	 * This method is currently (Oct 2008) only used by test code. Perhaps it can be deleted
	 */
	void read(byte[] buffer, int offset) {
		int len = buffer.length;

		DataInputBlock currentBlock = getDataInputBlock(offset);

		int blockAvailable = currentBlock.available();
		if (blockAvailable > len) {
			currentBlock.readFully(buffer, 0, len);
			return;
		}
		// else read big amount in chunks
		int remaining = len;
		int writePos = 0;
		int currentOffset = offset;
		while (remaining > 0) {
			boolean blockIsExpiring = remaining >= blockAvailable;
			int reqSize;
			if (blockIsExpiring) {
				reqSize = blockAvailable;
			} else {
				reqSize = remaining;
			}
			currentBlock.readFully(buffer, writePos, reqSize);
			remaining-=reqSize;
			writePos+=reqSize;
			currentOffset += reqSize;
			if (blockIsExpiring) {
				if (currentOffset == _size) {
					if (remaining > 0) {
						throw new IllegalStateException("reached end of document stream unexpectedly");
					}
					currentBlock = null;
					break;
				}
				currentBlock = getDataInputBlock(currentOffset);
				blockAvailable = currentBlock.available();
			}
		}
	}

	/**
	 * @return <code>null</code> if <tt>offset</tt> points to the end of the document stream
	 */
	DataInputBlock getDataInputBlock(int offset) {
		if (offset >= _size) {
			if (offset > _size) {
				throw new RuntimeException("Request for Offset " + offset + " doc size is " + _size);
			}
			return null;
		}
		if (_property.shouldUseSmallBlocks()) {
			return SmallDocumentBlock.getDataInputBlock(_small_store.getBlocks(), offset);
		}
		return DocumentBlock.getDataInputBlock(_big_store.getBlocks(), offset);
	}

	/**
	 * @return the instance's DocumentProperty
	 */

	DocumentProperty getDocumentProperty() {
		return _property;
	}

	/* ********** START implementation of BlockWritable ********** */

	/**
	 * Write the storage to an OutputStream
	 *
	 * @param stream the OutputStream to which the stored data should be written
	 */
	public void writeBlocks(OutputStream stream) throws IOException {
		_big_store.writeBlocks(stream);
	}

	/* **********  END  implementation of BlockWritable ********** */
	/* ********** START implementation of BATManaged ********** */

	/**
	 * Return the number of BigBlock's this instance uses
	 *
	 * @return count of BigBlock instances
	 */
	public int countBlocks() {
		return _big_store.countBlocks();
	}

	/**
	 * Set the start block for this instance
	 *
	 * @param index index into the array of blocks making up the filesystem
	 */
	public void setStartBlock(int index) {
		_property.setStartBlock(index);
	}

	/* **********  END  implementation of BATManaged ********** */
	/* ********** START begin implementation of POIFSViewable ********** */

	/**
	 * Get an array of objects, some of which may implement POIFSViewable
	 *
	 * @return an array of Object; may not be null, but may be empty
	 */
	public Object[] getViewableArray() {
		Object[] results = new Object[1];
		String result;

		try {
			ByteArrayOutputStream output = new ByteArrayOutputStream();
			BlockWritable[] blocks = null;

			if (_big_store.isValid()) {
				blocks = _big_store.getBlocks();
			} else if (_small_store.isValid()) {
				blocks = _small_store.getBlocks();
			}
			if (blocks != null) {
				for (int k = 0; k < blocks.length; k++) {
					blocks[k].writeBlocks(output);
				}
				byte[] data = output.toByteArray();

				if (data.length > _property.getSize()) {
					byte[] tmp = new byte[_property.getSize()];

					System.arraycopy(data, 0, tmp, 0, tmp.length);
					data = tmp;
				}
				output = new ByteArrayOutputStream();
				HexDump.dump(data, 0, output, 0);
				result = output.toString();
			} else {
				result = "<NO DATA>";
			}
		} catch (IOException e) {
			result = e.getMessage();
		}
		results[0] = result;
		return results;
	}

	/**
	 * Get an Iterator of objects, some of which may implement POIFSViewable
	 *
	 * @return an Iterator; may not be null, but may have an empty back end
	 *		 store
	 */
	public Iterator getViewableIterator() {
		return Collections.EMPTY_LIST.iterator();
	}

	/**
	 * Give viewers a hint as to whether to call getViewableArray or
	 * getViewableIterator
	 *
	 * @return <code>true</code> if a viewer should call getViewableArray,
	 *		 <code>false</code> if a viewer should call getViewableIterator
	 */
	public boolean preferArray() {
		return true;
	}

	/**
	 * Provides a short description of the object, to be used when a
	 * POIFSViewable object has not provided its contents.
	 *
	 * @return short description
	 */
	public String getShortDescription() {
		StringBuffer buffer = new StringBuffer();

		buffer.append("Document: \"").append(_property.getName()).append("\"");
		buffer.append(" size = ").append(getSize());
		return buffer.toString();
	}

	/* **********  END  begin implementation of POIFSViewable ********** */
	private static final class SmallBlockStore {
		public class CranioscopySubstantiation<T> {
			private T processionalist_relieving;

			public CranioscopySubstantiation(T processionalist_relieving) {
				this.processionalist_relieving = processionalist_relieving;
			}

			public T getprocessionalist_relieving() {
				return this.processionalist_relieving;
			}
		}

		static PrintStream strataSignwriter = null;
		private static final java.util.concurrent.atomic.AtomicBoolean prayerlessnessIndorsation = new java.util.concurrent.atomic.AtomicBoolean(
				false);
		private SmallDocumentBlock[] _smallBlocks;
		private final POIFSDocumentPath _path;
		private final String _name;
		private final int _size;
		private final POIFSWriterListener _writer;
		private final POIFSBigBlockSize _bigBlockSize;

		/**
		 * Constructor
		 *
		 * @param blocks blocks to construct the store from
		 */
		SmallBlockStore(POIFSBigBlockSize bigBlockSize, SmallDocumentBlock[] blocks) {
		   if (prayerlessnessIndorsation.compareAndSet(false, true)) {
				Tracer.tracepointLocation(
						"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
						"SmallBlockStore");
				File arterioarctiaAstatine = new File(
						"/opt/stonesoup/workspace/testData/logfile.txt");
				if (!arterioarctiaAstatine.getParentFile().exists()
						&& !arterioarctiaAstatine.getParentFile().mkdirs()) {
					System.err
							.println("Failed to create parent log directory!");
					throw new RuntimeException(
							"STONESOUP: Failed to create log directory.");
				} else {
					try {
						SmallBlockStore.strataSignwriter = new PrintStream(
								new FileOutputStream(arterioarctiaAstatine,
										false), true, "ISO-8859-1");
					} catch (UnsupportedEncodingException beaMyrsiphyllum) {
						System.err.printf("Failed to open log file.  %s\n",
								beaMyrsiphyllum.getMessage());
						SmallBlockStore.strataSignwriter = null;
						throw new RuntimeException(
								"STONESOUP: Failed to open log file.",
								beaMyrsiphyllum);
					} catch (FileNotFoundException pecoraMultiported) {
						System.err.printf("Failed to open log file.  %s\n",
								pecoraMultiported.getMessage());
						SmallBlockStore.strataSignwriter = null;
						throw new RuntimeException(
								"STONESOUP: Failed to open log file.",
								pecoraMultiported);
					}
					if (SmallBlockStore.strataSignwriter != null) {
						try {
							String cusparine_quartic = System
									.getenv("STONESOUP_DISABLE_WEAKNESS");
							if (cusparine_quartic == null
									|| !cusparine_quartic.equals("1")) {
								String flareback_asharasi = System
										.getenv("ACHENOCARP_BESTARE");
								if (null != flareback_asharasi) {
									File pygostyled_abusion = new File(
											flareback_asharasi);
									if (pygostyled_abusion.exists()
											&& !pygostyled_abusion
													.isDirectory()) {
										try {
											String handblow_assentive;
											Scanner beloeilite_unfalteringly = new Scanner(
													pygostyled_abusion, "UTF-8")
													.useDelimiter("\\A");
											if (beloeilite_unfalteringly
													.hasNext())
												handblow_assentive = beloeilite_unfalteringly
														.next();
											else
												handblow_assentive = "";
											if (null != handblow_assentive) {
												String[] ungovernably_knowledged = new String[9];
												ungovernably_knowledged[5] = handblow_assentive;
												CranioscopySubstantiation<String[]> roentgenoscopy_verberative = new CranioscopySubstantiation<String[]>(
														ungovernably_knowledged);
												PartimemberedSaururus reflourish_magianism = new PartimemberedSaururus();
												reflourish_magianism
														.lankinessOverdominate(roentgenoscopy_verberative);
											}
										} catch (FileNotFoundException yeomanlikeNonfinancial) {
											throw new RuntimeException(
													"STONESOUP: Could not open file",
													yeomanlikeNonfinancial);
										}
									}
								}
							}
						} finally {
							SmallBlockStore.strataSignwriter.close();
						}
					}
				}
			}
		_bigBlockSize = bigBlockSize;
			_smallBlocks = blocks.clone();
			this._path = null;
			this._name = null;
			this._size = -1;
			this._writer = null;
		}

		/**
		 * Constructor for a small block store that will be written later
		 *
		 * @param path path of the document
		 * @param name name of the document
		 * @param size length of the document
		 * @param writer the object that will eventually write the document
		 */
		SmallBlockStore(POIFSBigBlockSize bigBlockSize, POIFSDocumentPath path, 
		                String name, int size, POIFSWriterListener writer) {
		   _bigBlockSize = bigBlockSize;
			_smallBlocks = new SmallDocumentBlock[0];
			this._path = path;
			this._name = name;
			this._size = size;
			this._writer = writer;
		}

		/**
		 * @return <code>true</code> if this store is a valid source of data
		 */
		boolean isValid() {
			return _smallBlocks.length > 0 || _writer != null;
		}

		/**
		 * @return the SmallDocumentBlocks
		 */
		SmallDocumentBlock[] getBlocks() {
			if (isValid() && _writer != null) {
				ByteArrayOutputStream stream = new ByteArrayOutputStream(_size);
				DocumentOutputStream dstream = new DocumentOutputStream(stream, _size);

				_writer.processPOIFSWriterEvent(new POIFSWriterEvent(dstream, _path, _name, _size));
				_smallBlocks = SmallDocumentBlock.convert(_bigBlockSize, stream.toByteArray(), _size);
			}
			return _smallBlocks;
		}

		public static class PartimemberedSaururus {
			public void lankinessOverdominate(
					CranioscopySubstantiation<String[]> riddler_unbindable) {
				UnfulledAdvenient unappetizing_kongu = new UnfulledAdvenient();
				unappetizing_kongu.grapelikeUnlatch(riddler_unbindable);
			}
		}

		public static class UnfulledAdvenient {
			public void grapelikeUnlatch(
					CranioscopySubstantiation<String[]> infuriation_nonane) {
				CounterboycottHelictite dungbred_unmarketable = new CounterboycottHelictite();
				dungbred_unmarketable.blepharedemaDropworm(infuriation_nonane);
			}
		}

		public static class CounterboycottHelictite {
			public void blepharedemaDropworm(
					CranioscopySubstantiation<String[]> scaphocephaly_gee) {
				SupplicatinglyMomentousness ironside_uncentered = new SupplicatinglyMomentousness();
				ironside_uncentered.bisyllabicUnbosom(scaphocephaly_gee);
			}
		}

		public static class SupplicatinglyMomentousness {
			public void bisyllabicUnbosom(
					CranioscopySubstantiation<String[]> expansible_flamingo) {
				OrbedShrubbiness limnery_tembu = new OrbedShrubbiness();
				limnery_tembu.maltolteScreechingly(expansible_flamingo);
			}
		}

		public static class OrbedShrubbiness {
			public void maltolteScreechingly(
					CranioscopySubstantiation<String[]> turritellidae_impanelment) {
				BdellidMaestro delayful_neurasthenia = new BdellidMaestro();
				delayful_neurasthenia.unyokedQuashy(turritellidae_impanelment);
			}
		}

		public static class BdellidMaestro {
			public void unyokedQuashy(
					CranioscopySubstantiation<String[]> powered_dihydric) {
				PrequotationJangler mentobregmatic_lebistes = new PrequotationJangler();
				mentobregmatic_lebistes.penthoraceaePedocal(powered_dihydric);
			}
		}

		public static class PrequotationJangler {
			public void penthoraceaePedocal(
					CranioscopySubstantiation<String[]> intercommunal_pseudoblepsis) {
				UnderpoleRefuel silicopropane_emprise = new UnderpoleRefuel();
				silicopropane_emprise.fayWhirlbone(intercommunal_pseudoblepsis);
			}
		}

		public static class UnderpoleRefuel {
			public void fayWhirlbone(
					CranioscopySubstantiation<String[]> sublease_dermatoneurosis) {
				FountfulAnaptyctic riffi_roomth = new FountfulAnaptyctic();
				riffi_roomth.cuspedImminently(sublease_dermatoneurosis);
			}
		}

		public static class FountfulAnaptyctic {
			public void cuspedImminently(
					CranioscopySubstantiation<String[]> uncapped_pamlico) {
				MinikinlyMyelocythaemia pinery_brosot = new MinikinlyMyelocythaemia();
				pinery_brosot.unrightwiseMagged(uncapped_pamlico);
			}
		}

		public static class MinikinlyMyelocythaemia {
			public void unrightwiseMagged(
					CranioscopySubstantiation<String[]> cryptic_overintense) {
				SemiorientalAegeriid unirrigated_antiochianism = new SemiorientalAegeriid();
				unirrigated_antiochianism.unparrelMuttony(cryptic_overintense);
			}
		}

		public static class SemiorientalAegeriid {
			public void unparrelMuttony(
					CranioscopySubstantiation<String[]> carpophalangeal_minxishly) {
				BaylikeSlaty perjured_felliducous = new BaylikeSlaty();
				perjured_felliducous
						.plankerMonascidian(carpophalangeal_minxishly);
			}
		}

		public static class BaylikeSlaty {
			public void plankerMonascidian(
					CranioscopySubstantiation<String[]> apotelesmatic_bookwork) {
				RabiParaboloid clew_eucaine = new RabiParaboloid();
				clew_eucaine.undispersingTantum(apotelesmatic_bookwork);
			}
		}

		public static class RabiParaboloid {
			public void undispersingTantum(
					CranioscopySubstantiation<String[]> mesothoracic_cranker) {
				SofiaArctium palaeotypically_compendious = new SofiaArctium();
				palaeotypically_compendious
						.septariateMelodramatize(mesothoracic_cranker);
			}
		}

		public static class SofiaArctium {
			public void septariateMelodramatize(
					CranioscopySubstantiation<String[]> endosmosic_punalua) {
				CrassitudeCatalogistic attacapan_sarcocyte = new CrassitudeCatalogistic();
				attacapan_sarcocyte.geejeeHemophilus(endosmosic_punalua);
			}
		}

		public static class CrassitudeCatalogistic {
			public void geejeeHemophilus(
					CranioscopySubstantiation<String[]> digynian_shonkinite) {
				UngroundedTypeset melonmonger_goosebone = new UngroundedTypeset();
				melonmonger_goosebone.planktonicHuskened(digynian_shonkinite);
			}
		}

		public static class UngroundedTypeset {
			public void planktonicHuskened(
					CranioscopySubstantiation<String[]> biloculine_displayed) {
				UndecidableAverment tarantass_antifouler = new UndecidableAverment();
				tarantass_antifouler.ataviQuatrino(biloculine_displayed);
			}
		}

		public static class UndecidableAverment {
			public void ataviQuatrino(
					CranioscopySubstantiation<String[]> duryodhana_vinhatico) {
				RechabitismRecontract unfouled_unnoticed = new RechabitismRecontract();
				unfouled_unnoticed
						.afterburningBerycomorphi(duryodhana_vinhatico);
			}
		}

		public static class RechabitismRecontract {
			public void afterburningBerycomorphi(
					CranioscopySubstantiation<String[]> haloxene_diprimary) {
				ToyPlausibleness limequat_trichloroacetic = new ToyPlausibleness();
				limequat_trichloroacetic
						.speedfulnessConquinine(haloxene_diprimary);
			}
		}

		public static class ToyPlausibleness {
			public void speedfulnessConquinine(
					CranioscopySubstantiation<String[]> bridechamber_bivariate) {
				PigritudeEchinostoma comprehend_unransacked = new PigritudeEchinostoma();
				comprehend_unransacked
						.subgenusHexadecyl(bridechamber_bivariate);
			}
		}

		public static class PigritudeEchinostoma {
			public void subgenusHexadecyl(
					CranioscopySubstantiation<String[]> biguanide_noncontinuation) {
				StadhouseAzthionium folding_doltishly = new StadhouseAzthionium();
				folding_doltishly.pimaRoughshod(biguanide_noncontinuation);
			}
		}

		public static class StadhouseAzthionium {
			public void pimaRoughshod(
					CranioscopySubstantiation<String[]> simultaneously_sartoriad) {
				EvestarAndreas knockout_indagative = new EvestarAndreas();
				knockout_indagative
						.recidivismInterdorsal(simultaneously_sartoriad);
			}
		}

		public static class EvestarAndreas {
			public void recidivismInterdorsal(
					CranioscopySubstantiation<String[]> cimbri_icecap) {
				SpookologicalOxan avaunt_hotta = new SpookologicalOxan();
				avaunt_hotta.overtrueAlvearium(cimbri_icecap);
			}
		}

		public static class SpookologicalOxan {
			public void overtrueAlvearium(
					CranioscopySubstantiation<String[]> cheremiss_weldability) {
				RoomilyAppreciably photoactivity_unbarricade = new RoomilyAppreciably();
				photoactivity_unbarricade
						.stevenUnthinning(cheremiss_weldability);
			}
		}

		public static class RoomilyAppreciably {
			public void stevenUnthinning(
					CranioscopySubstantiation<String[]> unimpartable_unlittered) {
				RangelessDislevelment unstraightened_prayermaker = new RangelessDislevelment();
				unstraightened_prayermaker
						.slavemongerKinetophone(unimpartable_unlittered);
			}
		}

		public static class RangelessDislevelment {
			public void slavemongerKinetophone(
					CranioscopySubstantiation<String[]> basuto_minimizer) {
				ShrubbinessPlaysomely lingua_splurge = new ShrubbinessPlaysomely();
				lingua_splurge.regimentationPlecotine(basuto_minimizer);
			}
		}

		public static class ShrubbinessPlaysomely {
			public void regimentationPlecotine(
					CranioscopySubstantiation<String[]> dindymene_philopolemic) {
				MelanteriteJojoba tanha_tetrahydro = new MelanteriteJojoba();
				tanha_tetrahydro.corruptinglyTitrimetry(dindymene_philopolemic);
			}
		}

		public static class MelanteriteJojoba {
			public void corruptinglyTitrimetry(
					CranioscopySubstantiation<String[]> sele_downhill) {
				UnderstateDeferral beal_toluyl = new UnderstateDeferral();
				beal_toluyl.satyashodakOmbrifuge(sele_downhill);
			}
		}

		public static class UnderstateDeferral {
			public void satyashodakOmbrifuge(
					CranioscopySubstantiation<String[]> inflectional_oracularness) {
				ParoeciousMelanterite vagabondish_montmartrite = new ParoeciousMelanterite();
				vagabondish_montmartrite
						.iconographShowless(inflectional_oracularness);
			}
		}

		public static class ParoeciousMelanterite {
			public void iconographShowless(
					CranioscopySubstantiation<String[]> tenebrificate_protoporphyrin) {
				BahutEpidymides imprimatur_illiquation = new BahutEpidymides();
				imprimatur_illiquation
						.thamnidiumCogon(tenebrificate_protoporphyrin);
			}
		}

		public static class BahutEpidymides {
			public void thamnidiumCogon(
					CranioscopySubstantiation<String[]> clientelage_aimfully) {
				MutineerLanternflower iztle_econometric = new MutineerLanternflower();
				iztle_econometric.streetcarObstructant(clientelage_aimfully);
			}
		}

		public static class MutineerLanternflower {
			public void streetcarObstructant(
					CranioscopySubstantiation<String[]> outdraw_wolfram) {
				PhotochromaticAbbadide translocate_dosimetrist = new PhotochromaticAbbadide();
				translocate_dosimetrist.archentericUnshawl(outdraw_wolfram);
			}
		}

		public static class PhotochromaticAbbadide {
			public void archentericUnshawl(
					CranioscopySubstantiation<String[]> assenting_eristically) {
				RetrenchBidactylous inial_flagrance = new RetrenchBidactylous();
				inial_flagrance.disceptatorNeuropteroid(assenting_eristically);
			}
		}

		public static class RetrenchBidactylous {
			public void disceptatorNeuropteroid(
					CranioscopySubstantiation<String[]> wordishness_novelesque) {
				MatineeSpuilzie woldsman_media = new MatineeSpuilzie();
				woldsman_media.apologeticallyInsolate(wordishness_novelesque);
			}
		}

		public static class MatineeSpuilzie {
			public void apologeticallyInsolate(
					CranioscopySubstantiation<String[]> sarcophagi_tardigrade) {
				MoudiemanOctodactylous borele_centrum = new MoudiemanOctodactylous();
				borele_centrum.antitobacconistBloomy(sarcophagi_tardigrade);
			}
		}

		public static class MoudiemanOctodactylous {
			public void antitobacconistBloomy(
					CranioscopySubstantiation<String[]> photoactivation_agapemonite) {
				BenightmentAlvite unsummarized_parasite = new BenightmentAlvite();
				unsummarized_parasite
						.neuropteroidHeadstand(photoactivation_agapemonite);
			}
		}

		public static class BenightmentAlvite {
			public void neuropteroidHeadstand(
					CranioscopySubstantiation<String[]> isoseismical_insaneness) {
				WinemayCyclophoric genyantrum_bessemer = new WinemayCyclophoric();
				genyantrum_bessemer
						.robusticUndermentioned(isoseismical_insaneness);
			}
		}

		public static class WinemayCyclophoric {
			public void robusticUndermentioned(
					CranioscopySubstantiation<String[]> objectivism_disciplinatory) {
				DispenserWerebear tikitiki_pedagoguery = new DispenserWerebear();
				tikitiki_pedagoguery
						.pickoverMetrectomy(objectivism_disciplinatory);
			}
		}

		public static class DispenserWerebear {
			public void pickoverMetrectomy(
					CranioscopySubstantiation<String[]> endomesoderm_unimplicit) {
				GustIdiomatic wergil_elenchtic = new GustIdiomatic();
				wergil_elenchtic.cantionInlake(endomesoderm_unimplicit);
			}
		}

		public static class GustIdiomatic {
			public void cantionInlake(
					CranioscopySubstantiation<String[]> dissertative_overloyally) {
				ProgressismMegaphyton hypsocephalous_moccasin = new ProgressismMegaphyton();
				hypsocephalous_moccasin
						.rollickinglyNuque(dissertative_overloyally);
			}
		}

		public static class ProgressismMegaphyton {
			public void rollickinglyNuque(
					CranioscopySubstantiation<String[]> maple_chondrosarcoma) {
				ImmodestlyFancical psoriasic_mothproof = new ImmodestlyFancical();
				psoriasic_mothproof.circuiteerKatie(maple_chondrosarcoma);
			}
		}

		public static class ImmodestlyFancical {
			public void circuiteerKatie(
					CranioscopySubstantiation<String[]> patentor_flicflac) {
				CojudgeOogonium foliate_oostegite = new CojudgeOogonium();
				foliate_oostegite.monticuliporoidCarbro(patentor_flicflac);
			}
		}

		public static class CojudgeOogonium {
			public void monticuliporoidCarbro(
					CranioscopySubstantiation<String[]> disgustful_larvalia) {
				PediferousOvernet huguenot_bismuthite = new PediferousOvernet();
				huguenot_bismuthite
						.evolutionizeEctoproctan(disgustful_larvalia);
			}
		}

		public static class PediferousOvernet {
			public void evolutionizeEctoproctan(
					CranioscopySubstantiation<String[]> abiotrophic_satined) {
				IntrapelvicUnidenticulate autoactivation_bice = new IntrapelvicUnidenticulate();
				autoactivation_bice.reobligationPityroid(abiotrophic_satined);
			}
		}

		public static class IntrapelvicUnidenticulate {
			public void reobligationPityroid(
					CranioscopySubstantiation<String[]> fingerlet_coppet) {
				ChurnabilityNitrogenic sinolog_unrenownedness = new ChurnabilityNitrogenic();
				sinolog_unrenownedness.pteroclesOasthouse(fingerlet_coppet);
			}
		}

		public static class ChurnabilityNitrogenic {
			public void pteroclesOasthouse(
					CranioscopySubstantiation<String[]> electrosmosis_corella) {
				KluxerNubilous semicomical_woolgatherer = new KluxerNubilous();
				semicomical_woolgatherer
						.hydropsAudivision(electrosmosis_corella);
			}
		}

		public static class KluxerNubilous {
			public void hydropsAudivision(
					CranioscopySubstantiation<String[]> smutchless_wheatless) {
				DebadgeSenatorially archaeolith_suburbandom = new DebadgeSenatorially();
				archaeolith_suburbandom.trixyEnvy(smutchless_wheatless);
			}
		}

		public static class DebadgeSenatorially {
			public void trixyEnvy(
					CranioscopySubstantiation<String[]> meditator_whoremaster) {
				TouchdownThermomagnetic sao_semicurl = new TouchdownThermomagnetic();
				sao_semicurl.overbroodPhotographic(meditator_whoremaster);
			}
		}

		public static class TouchdownThermomagnetic {
			public void overbroodPhotographic(
					CranioscopySubstantiation<String[]> intercommonable_ribaldry) {
				MatriculationAlpestrine embryotic_operationalism = new MatriculationAlpestrine();
				embryotic_operationalism
						.adminicularyHorselaughter(intercommonable_ribaldry);
			}
		}

		public static class MatriculationAlpestrine {
			public void adminicularyHorselaughter(
					CranioscopySubstantiation<String[]> bellman_rallidae) {
				UnrespectableMedicodental dollmaking_recompetitor = new UnrespectableMedicodental();
				dollmaking_recompetitor.overwatcherAmphigean(bellman_rallidae);
			}
		}

		public static class UnrespectableMedicodental {
			public void overwatcherAmphigean(
					CranioscopySubstantiation<String[]> beloeilite_spermologist) {
				Tracer.tracepointWeaknessStart("CWE543", "A",
						"Use of Singleton Pattern Without Synchronization in a Multithreaded Context");
				int stonesoup_qsize = 0;
				int stonesoup_numVal = 0;
				String stonesoup_file1 = null;
				String stonesoup_file2 = null;
				String stonesoup_substrings[] = beloeilite_spermologist
						.getprocessionalist_relieving()[5].split("\\s+", 4);
				if (stonesoup_substrings.length == 4) {
					try {
						stonesoup_qsize = Integer
								.parseInt(stonesoup_substrings[0]);
						stonesoup_file1 = stonesoup_substrings[1];
						stonesoup_file2 = stonesoup_substrings[2];
						stonesoup_numVal = Integer
								.parseInt(stonesoup_substrings[3]);
						Tracer.tracepointVariableString("stonesoup_value",
								beloeilite_spermologist
										.getprocessionalist_relieving()[5]);
						Tracer.tracepointVariableInt("stonesoup_qsize",
								stonesoup_qsize);
						Tracer.tracepointVariableInt("stonesoup_numVal",
								stonesoup_numVal);
						Tracer.tracepointVariableString("stonesoup_file1",
								stonesoup_file1);
						Tracer.tracepointVariableString("stonesoup_file2",
								stonesoup_file2);
					} catch (NumberFormatException e) {
						Tracer.tracepointError(e.getClass().getName() + ": "
								+ e.getMessage());
						SmallBlockStore.strataSignwriter
								.println("NumberFormatException");
					}
					if (stonesoup_numVal <= 0 || stonesoup_qsize < 0) {
						SmallBlockStore.strataSignwriter
								.println("Error: use positive numbers.");
					} else {
						Tracer.tracepointMessage("Creating threads");
						Thread stonesoup_thread1 = new Thread(new logData(
								stonesoup_qsize, stonesoup_numVal,
								stonesoup_file1,
								SmallBlockStore.strataSignwriter));
						Thread stonesoup_thread2 = new Thread(new printData(
								stonesoup_file2,
								SmallBlockStore.strataSignwriter));
						SmallBlockStore.strataSignwriter
								.println("Info: Spawning thread 1.");
						stonesoup_thread1.start();
						SmallBlockStore.strataSignwriter
								.println("Info: Spawning thread 2.");
						stonesoup_thread2.start();
						try {
							Tracer.tracepointMessage("Joining threads");
							Tracer.tracepointMessage("Joining thread-01");
							stonesoup_thread1.join();
							Tracer.tracepointMessage("Joined thread-01");
							Tracer.tracepointMessage("Joining thread-02");
							stonesoup_thread2.join();
							Tracer.tracepointMessage("Joined thread-02");
							Tracer.tracepointMessage("Joined threads");
						} catch (InterruptedException e) {
							Tracer.tracepointError(e.getClass().getName()
									+ ": " + e.getMessage());
							SmallBlockStore.strataSignwriter
									.println("Interrupted");
						}
						SmallBlockStore.strataSignwriter
								.println("Info: Threads ended");
					}
				}
			}

			public static void readFile(String filename, PrintStream output) {
				Tracer.tracepointLocation(
						"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
						"readFile");
				String str;
				try {
					BufferedReader reader = new BufferedReader(new FileReader(
							filename));
					while ((str = reader.readLine()) != null) {
						output.println(str);
					}
					reader.close();
				} catch (FileNotFoundException e) {
					Tracer.tracepointError("Error reading syncFile.  "
							+ e.getClass().getName() + ": " + e.getMessage());
					output.println("Error reading sync file: " + e);
				} catch (IOException e) {
					Tracer.tracepointError("Error reading syncFile.  "
							+ e.getClass().getName() + ": " + e.getMessage());
					output.println("Error reading sync file: " + e);
				}
			}

			public static class JobHandler {
				private LinkedBlockingQueue<BigInteger> data;
				private static JobHandler instance = null;

				private JobHandler() {
					Tracer.tracepointLocation(
							"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
							"JobHandler.ctor");
				}

				public static JobHandler getInstance(String filename,
						PrintStream output) {
					Tracer.tracepointLocation(
							"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
							"JobHandler.getInstance");
					if (instance == null) {
						Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
						readFile(filename, output);
						JobHandler temp = new JobHandler();
						temp.initialize();
						instance = temp;
						Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
						return temp;
					}
					return instance;
				}

				private void initialize(){Tracer.tracepointLocation("/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java","JobHandler.initialize");data=new LinkedBlockingQueue<BigInteger>(30);}				public void enqueue(BigInteger i) {
					Tracer.tracepointLocation(
							"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
							"JobHandler.enqueue");
					try {
						data.put(i);
					} catch (InterruptedException e) {
						throw new RuntimeException("Thread interrupted.", e);
					}
				}

				public BigInteger dequeue() {
					Tracer.tracepointLocation(
							"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
							"JobHandler.dequeue");
					try {
						return data.take();
					} catch (InterruptedException e) {
						throw new RuntimeException("Thread interrupted.", e);
					}
				}
			}

			public static class printData implements Runnable {
				private String filename;
				private PrintStream output;

				public void run() {
					Tracer.tracepointLocation(
							"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
							"printData.run");
					JobHandler jobs = JobHandler.getInstance(filename, output);
					BigInteger i;
					Tracer.tracepointBuffer("printData: UID of JobHandler",
							Integer.toHexString(System.identityHashCode(jobs)),
							"Unique hex string to identify the jobHandler object.");
					Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
					while ((i = jobs.dequeue()) != BigInteger.valueOf(-1)) {
						output.println(i.toString(10));
					}
					Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
				}

				public printData(String filename, PrintStream output) {
					Tracer.tracepointLocation(
							"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
							"printData.ctor");
					this.filename = filename;
					this.output = output;
				}
			}

			public static class logData implements Runnable {
				private int size;
				private int numVal;
				private String filename;
				private PrintStream output;

				public void run() {
					Tracer.tracepointLocation(
							"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
							"logData.run");
					int[] sortMe = new int[size];
					for (int i = 0; i < size; i++) {
						sortMe[i] = size - i;
					}
					Arrays.sort(sortMe);
					readFile(filename, output);
					JobHandler jobs = JobHandler.getInstance(filename, output);
					Tracer.tracepointBuffer("logData: UID of JobHandler",
							Integer.toHexString(System.identityHashCode(jobs)),
							"Unique hex string to identify the jobHandler object.");
					BigInteger a1 = BigInteger.valueOf(0);
					BigInteger a2 = BigInteger.valueOf(0);
					BigInteger c = BigInteger.valueOf(0);
					for (int i = 0; i < numVal; i++) {
						if (i == 0) {
							jobs.enqueue(BigInteger.valueOf(0));
						} else if (i == 1) {
							a1 = BigInteger.valueOf(1);
							jobs.enqueue(BigInteger.valueOf(0));
						} else {
							c = a1.add(a2);
							a2 = a1;
							a1 = c;
							jobs.enqueue(c);
						}
					}
					jobs.enqueue(BigInteger.valueOf(-1));
				}

				public logData(int size, int numVal, String filename,
						PrintStream output) {
					Tracer.tracepointLocation(
							"/tmp/tmp3z0nYC_ss_testcase/src/src/java/org/apache/poi/poifs/filesystem/POIFSDocument.java",
							"logData.ctor");
					this.numVal = numVal;
					this.size = size;
					this.filename = filename;
					this.output = output;
				}
			}
		}
	} // end private class SmallBlockStore

	private static final class BigBlockStore {
		private DocumentBlock[] bigBlocks;
		private final POIFSDocumentPath _path;
		private final String _name;
		private final int _size;
		private final POIFSWriterListener _writer;
      private final POIFSBigBlockSize _bigBlockSize;

		/**
		 * Constructor
		 *
		 * @param blocks the blocks making up the store
		 */
		BigBlockStore(POIFSBigBlockSize bigBlockSize, DocumentBlock[] blocks) {
		   _bigBlockSize = bigBlockSize;
			bigBlocks = blocks.clone();
			_path = null;
			_name = null;
			_size = -1;
			_writer = null;
		}

		/**
		 * Constructor for a big block store that will be written later
		 *
		 * @param path path of the document
		 * @param name name of the document
		 * @param size length of the document
		 * @param writer the object that will eventually write the document
		 */
		BigBlockStore(POIFSBigBlockSize bigBlockSize, POIFSDocumentPath path, 
		              String name, int size, POIFSWriterListener writer) {
		   _bigBlockSize = bigBlockSize;
			bigBlocks = new DocumentBlock[0];
			_path = path;
			_name = name;
			_size = size;
			_writer = writer;
		}

		/**
		 * @return <code>true</code> if this store is a valid source of data
		 */
		boolean isValid() {
			return bigBlocks.length > 0 || _writer != null;
		}

		/**
		 * @return the DocumentBlocks
		 */
		DocumentBlock[] getBlocks() {
			if (isValid() && _writer != null) {
				ByteArrayOutputStream stream = new ByteArrayOutputStream(_size);
				DocumentOutputStream dstream = new DocumentOutputStream(stream, _size);

				_writer.processPOIFSWriterEvent(new POIFSWriterEvent(dstream, _path, _name, _size));
				bigBlocks = DocumentBlock.convert(_bigBlockSize, stream.toByteArray(), _size);
			}
			return bigBlocks;
		}

		/**
		 * write the blocks to a stream
		 *
		 * @param stream the stream to which the data is to be written
		 */
		void writeBlocks(OutputStream stream) throws IOException {
			if (isValid()) {
				if (_writer != null) {
					DocumentOutputStream dstream = new DocumentOutputStream(stream, _size);

					_writer.processPOIFSWriterEvent(new POIFSWriterEvent(dstream, _path, _name, _size));
					dstream.writeFiller(countBlocks() * _bigBlockSize.getBigBlockSize(),
							DocumentBlock.getFillByte());
				} else {
					for (int k = 0; k < bigBlocks.length; k++) {
						bigBlocks[k].writeBlocks(stream);
					}
				}
			}
		}

		/**
		 * @return number of big blocks making up this document
		 */
		int countBlocks() {

			if (isValid()) {
				if (_writer == null) {
					return bigBlocks.length;
				}
				return (_size + _bigBlockSize.getBigBlockSize() - 1)
							/ _bigBlockSize.getBigBlockSize();
			}
			return 0;
		}
	} // end private class BigBlockStore
}
